﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.Common;
using Microsoft.TeamFoundation.VersionControl.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.TeamFoundation;
using Microsoft.VisualStudio.TeamFoundation.VersionControl;
using Microsoft.VisualStudio.TeamFoundation.WorkItemTracking;

namespace TFSExt.DestroyWorkItems
{
    /// <summary>
    /// This is the class that implements the package exposed by this assembly.
    ///
    /// The minimum requirement for a class to be considered a valid package for Visual Studio
    /// is to implement the IVsPackage interface and register itself with the shell.
    /// This package uses the helper classes defined inside the Managed Package Framework (MPF)
    /// to do it: it derives from the Package class that provides the implementation of the 
    /// IVsPackage interface and uses the registration attributes defined in the framework to 
    /// register itself and its components with the shell.
    /// </summary>
    [PackageRegistration(UseManagedResourcesOnly = true)]
    [InstalledProductRegistration("#110", "#112", "1.0", IconResourceID = 400)]
    [ProvideLoadKey("Standard", "1.0", "TFS Productivity Tools - Destroy Work Items", "Mark Babayev (Experis)", 104)]
    [ProvideMenuResource("Menus.ctmenu", 1)]
    [Guid(GuidList.guidDestroyWorkItemsPkgString)]
    public sealed class DestroyWorkItemsPackage : Package
    {
        private EnvDTE.DTE dte = null;
        private TfsTeamProjectCollection tfscoll = null;
        private VersionControlServer vcsrv = null;
        private VersionControlExt vcext = null;
        private string AppTitle = null;
        
        /// <summary>
        /// Default constructor of the package.
        /// Inside this method you can place any initialization code that does not require 
        /// any Visual Studio service because at this point the package object is created but 
        /// not sited yet inside Visual Studio environment. The place to do all the other 
        /// initialization is the Initialize method.
        /// </summary>
        public DestroyWorkItemsPackage()
        {
            Trace.WriteLine(string.Format(CultureInfo.CurrentCulture, "Entering constructor for: {0}", this.ToString()));
        }

        /// <summary>
        /// Initializes variables by extension services.
        /// </summary>
        private void InitializeExtVariables()
        {
            dte = (EnvDTE.DTE)GetService(typeof(EnvDTE.DTE));
            //IVsUIShell uiShell = (IVsUIShell)GetService(typeof(SVsUIShell));

            var TFSExt = dte.GetObject("Microsoft.VisualStudio.TeamFoundation.TeamFoundationServerExt") as TeamFoundationServerExt;
            vcext = dte.GetObject("Microsoft.VisualStudio.TeamFoundation.VersionControl.VersionControlExt") as VersionControlExt;

            //tfscoll = new TfsTeamProjectCollection(new Uri(TFSExt.ActiveProjectContext.DomainUri));
            tfscoll = TfsTeamProjectCollectionFactory.GetTeamProjectCollection(new Uri(TFSExt.ActiveProjectContext.DomainUri));
            tfscoll.Connect(Microsoft.TeamFoundation.Framework.Common.ConnectOptions.None);
            tfscoll.EnsureAuthenticated();

            vcsrv = (VersionControlServer)tfscoll.GetService(typeof(VersionControlServer));

            AppTitle = Resources.ResourceManager.GetString("AppTitle");

            var outputWindow = GetService(typeof(SVsOutputWindow)) as IVsOutputWindow;
            var dialogFactory = GetService(typeof(SVsThreadedWaitDialogFactory)) as IVsThreadedWaitDialogFactory;

            Utilities.dialogFactory = dialogFactory;
            Utilities.outputWindow = outputWindow;
            Utilities.AppTitle = AppTitle;
        }

        /////////////////////////////////////////////////////////////////////////////
        // Overriden Package Implementation
        #region Package Members

        /// <summary>
        /// Initialization of the package; this method is called right after the package is sited, so this is the place
        /// where you can put all the initilaization code that rely on services provided by VisualStudio.
        /// </summary>
        protected override void Initialize()
        {
            Trace.WriteLine (string.Format(CultureInfo.CurrentCulture, "Entering Initialize() of: {0}", this.ToString()));
            base.Initialize();

            // Add our command handlers for menu (commands must exist in the .vsct file)
            OleMenuCommandService mcs = GetService(typeof(IMenuCommandService)) as OleMenuCommandService;
            if ( null != mcs )
            {
                // Create the command for the menu item.
                CommandID menuCommandID = new CommandID(GuidList.guidDestroyWorkItemsCmdSet, (int)PkgCmdIDList.cmdidMyCommand);
                MenuCommand menuItem = new MenuCommand(MenuItemCallback, menuCommandID );
                mcs.AddCommand( menuItem );
            }
        }
        #endregion

        /// <summary>
        /// This function is the callback used to execute a command when the a menu item is clicked.
        /// See the Initialize method to see how the menu item is associated to this function using
        /// the OleMenuCommandService service and the MenuCommand class.
        /// </summary>
        private void MenuItemCallback(object sender, EventArgs e)
        {
            InitializeExtVariables();
            var origCursor = Cursor.Current;
            Cursor.Current = Cursors.WaitCursor;

            DocumentService docsrv2 = (DocumentService)GetService(typeof(DocumentService));
            WorkItemStore wistore = (WorkItemStore)tfscoll.GetService(typeof(WorkItemStore));

            object _lockToken1 = new object();
            object _lockToken2 = new object();

            IWorkItemTrackingDocument doc2 = docsrv2.FindDocument(dte.ActiveDocument.FullName, _lockToken2);
            if (doc2 == null) return;

            var wicoll = new List<int>();

            foreach (int i in (doc2 as IResultsDocument).SelectedItemIds)
            {
                IWorkItemDocument widoc = docsrv2.GetWorkItem(tfscoll, i, _lockToken1);
                if (widoc == null) continue;
                widoc.Load();

                wicoll.Add(widoc.Item.Id);

                Utilities.OutputCommandString("Preparing for destroy: " + widoc.Item.Id + ", " + widoc.Item.Title);

                widoc.Release(_lockToken1);
            }
            doc2.Release(_lockToken2);

            foreach (var result in wistore.DestroyWorkItems(wicoll))
            {
                Utilities.OutputCommandString("Exception destroying Work Item #"+result.Id+" :"+result.Exception.Message);
            }

            Cursor.Current = origCursor;

            if (wicoll.Count > 0) 
                MessageBox.Show("Destroyed " + wicoll.Count + " Work Items", AppTitle, MessageBoxButtons.OK, MessageBoxIcon.Information);
        }
    }
}
