﻿using System;
using System.Diagnostics;
using System.Globalization;
using System.Runtime.InteropServices;
using System.ComponentModel.Design;
using Microsoft.Win32;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.OLE.Interop;
using Microsoft.VisualStudio.Shell;
using Microsoft.TeamFoundation.VersionControl.Client;
using Microsoft.TeamFoundation.Client;
using System.IO;
using System.Text;
using System.Linq;
using Microsoft.VisualStudio.TeamFoundation;
using Microsoft.VisualStudio.TeamFoundation.VersionControl;
using System.Windows.Forms;

namespace TFSExt.GetPreview
{
    /// <summary>
    /// This is the class that implements the package exposed by this assembly.
    ///
    /// The minimum requirement for a class to be considered a valid package for Visual Studio
    /// is to implement the IVsPackage interface and register itself with the shell.
    /// This package uses the helper classes defined inside the Managed Package Framework (MPF)
    /// to do it: it derives from the Package class that provides the implementation of the 
    /// IVsPackage interface and uses the registration attributes defined in the framework to 
    /// register itself and its components with the shell.
    /// </summary>
    // This attribute tells the PkgDef creation utility (CreatePkgDef.exe) that this class is
    // a package.
    [PackageRegistration(UseManagedResourcesOnly = true)]
    // This attribute is used to register the informations needed to show the this package
    // in the Help/About dialog of Visual Studio.
    [InstalledProductRegistration("#110", "#112", "1.0", IconResourceID = 400)]
    // This attribute is needed to let the shell know that this package exposes some menus.
    [ProvideMenuResource("Menus.ctmenu", 1)]
    [Guid(GuidList.guidGetPreviewPkgString)]
    [ProvideLoadKey("Standard", "1.0", "TFS Productivity Tools - Get Preview", "Mark Babayev (Experis)", 114)]
    public sealed class GetPreviewPackage : Package
    {
        string lastFolderDisplayed = null;
        string caption = null;
        
        /// <summary>
        /// Default constructor of the package.
        /// Inside this method you can place any initialization code that does not require 
        /// any Visual Studio service because at this point the package object is created but 
        /// not sited yet inside Visual Studio environment. The place to do all the other 
        /// initialization is the Initialize method.
        /// </summary>
        public GetPreviewPackage()
        {
            Trace.WriteLine(string.Format(CultureInfo.CurrentCulture, "Entering constructor for: {0}", this.ToString()));
        }



        /////////////////////////////////////////////////////////////////////////////
        // Overriden Package Implementation
        #region Package Members

        /// <summary>
        /// Initialization of the package; this method is called right after the package is sited, so this is the place
        /// where you can put all the initilaization code that rely on services provided by VisualStudio.
        /// </summary>
        protected override void Initialize()
        {
            Trace.WriteLine (string.Format(CultureInfo.CurrentCulture, "Entering Initialize() of: {0}", this.ToString()));
            base.Initialize();

            // Add our command handlers for menu (commands must exist in the .vsct file)
            OleMenuCommandService mcs = GetService(typeof(IMenuCommandService)) as OleMenuCommandService;
            if ( null != mcs )
            {
                // Create the command for the menu item.
                CommandID menuCommandID = new CommandID(GuidList.guidGetPreviewCmdSet, (int)PkgCmdIDList.cmdidMyCommand);
                MenuCommand menuItem = new MenuCommand(MenuItemCallback, menuCommandID );
                mcs.AddCommand( menuItem );
            }
        }
        #endregion

        /// <summary>
        /// This function is the callback used to execute a command when the a menu item is clicked.
        /// See the Initialize method to see how the menu item is associated to this function using
        /// the OleMenuCommandService service and the MenuCommand class.
        /// </summary>
        private void MenuItemCallback(object sender, EventArgs e)
        {
            var dte = (EnvDTE.DTE)GetService(typeof(EnvDTE.DTE));
            var TFSExt = dte.GetObject("Microsoft.VisualStudio.TeamFoundation.TeamFoundationServerExt") as TeamFoundationServerExt;
            var vcext = dte.GetObject("Microsoft.VisualStudio.TeamFoundation.VersionControl.VersionControlExt") as VersionControlExt;

            var tfscoll = TfsTeamProjectCollectionFactory.GetTeamProjectCollection(new Uri(TFSExt.ActiveProjectContext.DomainUri));
            tfscoll.Connect(Microsoft.TeamFoundation.Framework.Common.ConnectOptions.None);
            tfscoll.EnsureAuthenticated();

            var vcServer = (VersionControlServer)tfscoll.GetService(typeof(VersionControlServer));
            Workspace myWorkspace = vcext.Explorer.Workspace;

            caption = Resources.ResourceManager.GetString("AppTitle");

            var origCursor = Cursor.Current;
            Cursor.Current = Cursors.WaitCursor;

            vcServer.Getting += new GettingEventHandler(vcServer_Getting);

            foreach (string sourcePath in vcext.Explorer.SelectedItems.Select(x => x.SourceServerPath))
            {
                PrintOutputWindow("Previewing GET command for path: " + sourcePath);
                var getreq = new GetRequest(new ItemSpec(sourcePath, RecursionType.Full), VersionSpec.Latest);
                var status = myWorkspace.Get(getreq, GetOptions.Preview);

                PrintOutputWindow("HaveResolvableWarnings=" + status.HaveResolvableWarnings);
                PrintOutputWindow("NumConflicts=" + status.NumConflicts);
                PrintOutputWindow("NumFailures=" + status.NumFailures);
                PrintOutputWindow("NumOperations=" + status.NumOperations);
                PrintOutputWindow("NumUpdated=" + status.NumUpdated);
                PrintOutputWindow("NumWarnings=" + status.NumWarnings);
            }

            vcServer.Getting -= new GettingEventHandler(vcServer_Getting);

            Cursor.Current = origCursor;
        }

        private void PrintOutputWindow(string text)
        {
            var outputText = new StringBuilder();
            outputText.AppendFormat("{0}: {1} ", caption, text);
            outputText.AppendLine();
            // --- Get a reference to IVsOutputWindow. 
            var outputWindow = GetService(typeof(SVsOutputWindow)) as IVsOutputWindow;
            if (outputWindow == null) return;

            // --- Get the window pane for the general output. 
            var guidGeneral = Microsoft.VisualStudio.VSConstants.GUID_OutWindowDebugPane;
            IVsOutputWindowPane windowPane;

            if (Microsoft.VisualStudio.ErrorHandler.Failed(outputWindow.GetPane(ref guidGeneral, out windowPane)))
            {
                return;
            }
            // --- As the last step, write to the output window pane 
            windowPane.OutputString(outputText.ToString());
            windowPane.Activate();
        }

        void vcServer_Getting(object sender, GettingEventArgs e)
        {
            string fileName = null;
            if (e.TargetLocalItem != null)
            {
                string dirName = Path.GetDirectoryName(e.TargetLocalItem);
                fileName = Path.GetFileName(e.TargetLocalItem);

                if (e.TargetLocalItem == dirName)
                {
                    fileName = e.TargetLocalItem;
                }
                else
                {
                    if (lastFolderDisplayed != dirName)
                    {
                        if (lastFolderDisplayed != null)
                        {
                            PrintOutputWindow("");
                        }
                        PrintOutputWindow(dirName);
                        lastFolderDisplayed = dirName;
                    }
                }
            }

            string str;
            string message = e.GetMessage(fileName, out str);
            switch (e.Status)
            {
                case OperationStatus.Conflict:
                case OperationStatus.SourceWritable:
                case OperationStatus.TargetLocalPending:
                case OperationStatus.TargetWritable:
                case OperationStatus.SourceDirectoryNotEmpty:
                case OperationStatus.TargetIsDirectory:
                case OperationStatus.UnableToRefresh:
                    PrintOutputWindow("Warning: " + str);
                    return;

                case OperationStatus.Getting:
                case OperationStatus.Replacing:
                case OperationStatus.Deleting:
                    PrintOutputWindow(message);
                    return;
            }
        }
    }
}
